"""This module contains enums for various database operations and states."""

from enum import Enum
from enum import auto


class SQLOperation(str, Enum):
    """
    Enum for SQL operation types.

    Specifies the types of SQL operations that can be performed.
    """

    PROCEDURE = "procedure"  # A stored procedure operation.
    FUNCTION = "function"  # A function call operation.
    QUERY = "query"  # A standard SQL query operation.


class Dialect(str, Enum):
    """
    Enum for supported database dialects.

    Defines the database dialects supported by the application.
    """

    MYSQL = "mysql"  # MySQL database dialect.
    MARIADB = "mariadb"  # MariaDB database dialect.
    MSSQL = "mssql"  # Microsoft SQL Server dialect.
    POSTGRESQL = "postgres"  # postgres database dialect.
    ORACLE = "oracle"  # Oracle database dialect.
    SQLITE = "sqlite"  # SQLite database dialect.


class Handlers(str, Enum):
    """
    Enum for handler types.
    """

    ERROR_HANDLER = "error_handler"
    EXECUTION_HANDLER = "execution_handler"
    parameter_handlers = "parameter_handlers"
    STATEMENT_HANDLERS = "statement_handlers"
    RESULT_HANDLERS = "result_handlers"


class ExecutionState(str, Enum):
    """
    Enum for execution states.

    Represents the current state of database execution operations.
    """

    IDLE = "idle"  # Execution is idle.
    BEGIN = "begin"  # Execution is active.
    PROCESS_PARAMS = "process_params"  # Processing parameters.
    ADD_PAGING = "add_paging"  # Paging is being added to the query.
    FETCH = "fetch"  # Fetching results from the database.
    EXECUTE = "execute"  # Execution is in progress.
    PROCESS_RESULTS = "process_results"  # Processing results.
    COMMIT = "commit"  # Commiting changes.
    ERROR = "error"  # Execution encountered an error.


class ConnectionState(Enum):
    """
    Enum for connection states.

    Tracks the status of the database connection.
    """

    CONNECTED = auto()  # Connection is active.
    DISCONNECTED = auto()  # Connection is not active.
    ERROR = auto()  # Connection encountered an error.B


class ResultKeys(str, Enum):
    """
    Enum for result keys.
    """

    FETCHALL = "fetchall"
    FETCHONE = "fetchone"
    FETCHMANY = "fetchmany"
    ALL = "all"
    SCALARS = "scalars"
    COLUMNS = "columns"
    MAPPINGS = "mappings"
    RETURN_ROWS = "return_rows"
    ROWCOUNT = "rowcount"
    KEYS = "keys"



"""
Database Validation Models used for validating configurations
"""

# pyright: reportArgumentType=false
# pylint: disable=broad-except, W0707
# ruff: noqa: TC001
from __future__ import annotations

from typing import Any
from typing import Literal
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator
from pydantic_core import PydanticCustomError
from sqlalchemy.engine.url import make_url
from typing_extensions import Self
from elixirdb.enums import Dialect
from elixirdb.models.model import StrictModel
from elixirdb.models.options import EngineOptions
from elixirdb.models.options import SessionOptions
from elixirdb.models.urls import engines_url


driver_map = {
    Dialect.MYSQL: "mysql+pymysql",
    Dialect.MARIADB: "mariadb+pymysql",
    Dialect.MSSQL: "mssql+pymssql",
    Dialect.POSTGRESQL: "postgresql+psycopg2",
    Dialect.ORACLE: "oracle+oracledb",
    Dialect.SQLITE: "sqlite",
}


class UrlParams(StrictModel):
    """
    Url Connection parameters for the database.

    This model is used to store the necessary information to connect to
    a database.The driver is optional only if odbc is not utilized. If
    url is present in the model, it will override connection parameters.
    Review the model_validator in EngineModel for more information.

    """

    drivername: str = Field(
        "",
        description=(
            "SQLAlchemy database driver identifier. It may be a single dialect "
            "or a dialect and driver combination separated by '+'. "
        ),
        examples=["mysql+pymysql", "postgresql+psycopg2", "oracle+cx_oracle"],
    )
    host: str
    port: str = ""
    username: str = ""
    password: str = ""
    database: str = ""
    query: dict = Field(
        default_factory=dict,
        description="Query parameters for the URL string. (e.g. url/?key=value)",
    )

    @field_validator("port", mode="before")
    @classmethod
    def validate_port(cls, value: Any) -> str:
        """
        Will coerce a port to string.
        """
        return str(value)


class Statements(StrictModel):
    """
    Query model for database operations. This is experimental at the moment
    and the uses aren't fully flushed out.

    """

    schema_name: str | None = None
    prefix_raw_statements: bool = Field(
        False,
        description="Prefix tables defined in raw sql statements. Not in use",
    )
    prefix_procedures: bool = Field(
        default=False,
        description=(
            "When executing stored procedures, automatically prefix the "
            "procedure name with the schema defined in 'schema'."
        ),
    )

    def __init__(self, **data):
        """Sets the default values when enabling schema."""

        super().__init__(**data)
        p_queries = data.get("prefix_raw_statements", None)
        p_procedures = data.get("prefix_procedures", None)

        # Set the defaults to be True if not defined
        if self.schema_name and (p_queries is None and p_procedures is None):
            self.prefix_raw_statements = True
            self.prefix_procedures = True

    @model_validator(mode="before")
    @classmethod
    def validate_statements(cls, values: dict) -> dict:
        """
        Ensure correct values required by schema is set.

        Check before to ensure the user isn't passing the False value themselves.

        """

        schema_prefix = values.get("schema_name")
        prefix_raw_statements = values.get("prefix_raw_statements", None)
        prefix_procedures = values.get("prefix_procedures", None)

        if (
            schema_prefix
            and prefix_raw_statements is False
            and prefix_procedures is False
        ):
            raise ValueError(
                "Schema_prefix is provided, but both prefix_raw_statements and "
                "prefix_procedures are disabled."
            )

        if not schema_prefix and (prefix_raw_statements or prefix_procedures):
            raise ValueError(
                "schema_prefix is not provided, but prefix_raw_statements "
                "or prefix_procedures is set to 'True'"
            )

        return values


class EngineModel(BaseModel):
    """
    Single sqlalchemy database engine configuration model.

    The model forces a configuration to align with the expected schema
    to ensure that the database connection has the necessary parameters
    to create a connection.

    """

    debug: bool = False

    name: str = Field("", description="Name of the engine. Not in use.")
    dialect: Literal["mysql", "postgres", "mariadb", "sqlite", "mssql", "oracle"]
    default: bool = Field(
        False,
        description=(
            "Set the current engine as the default engine "
            "in a EnginerManager Model. (Multiple engines defined)"
        ),
    )
    auto_connect: bool = True
    apply_textclause: bool = Field(
        True,
        description=(
            "Apply textclause to any query sent to execute as a raq string."
        ),
        examples=["'SELECT 1' -> textclause('SELECT 1')"],
    )
    url: str | None = Field(
        None,
        description=("SQLAlchemy database connection url string."),
    )
    url_params: UrlParams | None = Field(
        None,
        description="URL parameters. Can be used instead of url. Useful when "
        "there are characters in the password that need escaping",
    )

    result_to_dict: bool = Field(
        True,
        description="Return results as dict. Only used in fetch_results method.",
    )

    session_options: SessionOptions | None = Field(
        None,
        description="See :class:`SessionOptions` for more information",
    )

    engine_options: EngineOptions | None = Field(
        None,
        description="Options passed into create_engine. See :class:`EngineOptions`",
    )
    statements: Statements = Field(default_factory=Statements)
    meta: dict[str, Any] = Field(
        default_factory=dict, description="Unused dictionary field. "
    )

    model_config = ConfigDict(extra="ignore", use_enum_values=True)

    @field_validator("url", mode="before")
    @classmethod
    def validate_sqlalchemy_url(cls, value: str | None) -> str | None:
        """
        Validates the provided url attribute.
        """
        if value is None:
            return value
        try:
            # Use SQLAlchemy's make_url to validate the URL
            make_url(value)
        except Exception:
            # raise ValueError("Url must match SqlAlchemy's engine url construct.")
            raise PydanticCustomError(  # noqa: B904
                "value_error",
                f"Invalid url format. Value provided: {value}",
                {"url": f"{engines_url}#sqlalchemy.engine.make_url"},
            )
        return value

    @model_validator(mode="after")
    def check_if_both_url_and_params(self) -> Self:
        """
        Checks if both `url` and `url_params` are provided. Both
        are mutually exclusive. (For this version)
        """
        if self.url and self.url_params:
            raise PydanticCustomError(
                "user_error",
                "Config should have either url or connection params, not both.",
            )
        return self

    @model_validator(mode="after")
    def check_if_not_url_and_not_params(self) -> Self:
        """
        Checks if both `url` and `url_params` are provided.
        """

        if not self.url and not self.url_params:
            raise PydanticCustomError(
                "user_error",
                "Each config requires a url or url_params.",
            )
        return self

    @model_validator(mode="after")
    def assign_driver(self) -> Self:
        """
        Assigns a default driver if not provided.
        """

        if self.url_params and not self.url_params.drivername:
            self.url_params.drivername = driver_map[self.dialect]
        return self



# SPDX-FileCopyrightText: 2024-present Victor Nguyen <victor.win86@gmail.com>
#
# SPDX-License-Identifier: MIT
# ruff: noqa: F401
# flake8: noqa: E501
# pyright: ignore[reportUnusedImport, W0404]
# pylint: disable=W0404
from __future__ import annotations

from elixirdb.db import ElixirDB
from elixirdb.db import ElixirDBStatements
from elixirdb.db import StatementsMixin
from elixirdb.db import create_db
from elixirdb.exc import print_and_raise_validation_errors
from elixirdb.models.engine import EngineModel
from elixirdb.models.manager import EngineManager
from elixirdb.models.options import EngineOptions
from elixirdb.models.options import ExecutionOptions
from elixirdb.models.options import SessionOptions
from elixirdb.utils.files import load_config
from elixirdb.utils.files import scan_files


__all__ = [
    "ElixirDB",
    "ElixirDBStatements",
    "EngineManager",
    "EngineModel",
    "EngineOptions",
    "ExecutionOptions",
    "SessionOptions",
    "StatementsMixin",
    "create_db",
    "load_config",
    "print_and_raise_validation_errors",
    "scan_files",
]



"""
Database connection and configuration management.

"""

# ruff: noqa: D102, PLR0916
# flake8: noqa: E501,
# pylint: disable=C0103,C0301,R0912,W0212
# pyright: reportUnknownVariableType=false, reportAttributeAccessIssue=false
from __future__ import annotations

import warnings
from functools import wraps
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Literal
from typing import Sequence
from sqlalchemy import CursorResult
from sqlalchemy import Executable
from sqlalchemy import create_engine
from sqlalchemy import text
from sqlalchemy.orm import scoped_session
from sqlalchemy.orm import sessionmaker
from typing_extensions import Self
from elixirdb.base import ConnectionBase
from elixirdb.base import CursorResultMetadata
from elixirdb.enums import ConnectionState
from elixirdb.enums import ExecutionState
from elixirdb.exc import CursorResultError
from elixirdb.exc import EngineKeyNotFoundError
from elixirdb.exc import InvalidElixirConfigError
from elixirdb.exc import InvalidEngineTypeError
from elixirdb.exc import NoSessionFactoryError
from elixirdb.handlers import handler as h_
from elixirdb.models.manager import EngineModel
from elixirdb.utils.db_utils import apply_schema_to_statement


if TYPE_CHECKING:
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams
    from sqlalchemy.engine.result import RowMapping
    from sqlalchemy.orm.session import Session
    from elixirdb.types import DatabaseEngineConfig
    from elixirdb.types import EngineType
    from elixirdb.types import QueryResult
    from elixirdb.types import RowData


class ElixirDB(ConnectionBase):
    """Create a SqlAlchemyDatabase object from a configuration file."""

    # Provide a scopefunc for the session factory.
    scopefunc: Callable[[], Any] | None = None

    def __init__(
        self,
        config: DatabaseEngineConfig | None = None,
        engine_key: str | None = None,
        engine_type: EngineType = "direct",
        session_factory: scoped_session[Session]
        | sessionmaker[Session]
        | None = None,
        **kwargs,
    ):
        """
        Create a connection manager for SqlAlchemy Connections using a mapping
        such as a configuration file. This library provides support automatically
        for yaml files if the configuration has the prefix `elixir` within it.

        e.g. `elixir_db_config.yaml`
        e.g. `myelixir_config.yaml`

        A typed dict has been provided as well as the model to which it validates.

        For typedDict see :class:`elixirdb.models.manager.DatabaseEngineConfig`
        For model see :class:`elixirdb.models.manager.DatabaseEngineConfig`

        Args:
            config (DatabaseEngineConfig | None): The configuration used to initialize

            engine_key (str | None): Identifier for multi-db setups.
            engine_type (EngineType): Mode for session management:
                direct: Raw connection.
                scoped: Thread-local sessions.
                session: Standard session.
            session_factory (SessionType | None): Pre-configured session factory.
            **kwargs: Additional connection parameters.

        Raises:
            ModelValidationError: If configuration validation fails.
            InvalidConfigModel: If engine_key conflicts with single config.
            InvalidEngineTypeError: If engine_type is invalid.
        """
        super().__init__(config=config, engine_key=engine_key, **kwargs)

        self.debug = self.db.debug
        self.engine_type = engine_type or "session"
        self.session_factory = session_factory

        if engine_type != "direct":
            if not self.session_factory:
                self.session_factory = self.create_session_factory(
                    engine=self.engine, engine_type=engine_type
                )
            if engine_type == "scoped":
                self.session = self.session_factory
        if self.db.auto_connect and engine_type != "scoped":
            self.connect()

    def __repr__(self) -> str:
        if isinstance(self.db, EngineModel):
            config = self.db.model_dump(exclude_unset=True) if self.db else ""
        else:
            config = self.db
        # dialect = self.db.dialect if self.db else ""
        return f"Model={self.__class__.__name__}, config={config}>"

    def __enter__(self) -> Self:
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit context"""
        if exc_type is not None and self.connection is not None:
            self.connection.rollback()
        self.close()

    def __getattr__(self, name: str) -> Any:
        """
        Wrap attribute access with parameter and result handling.

        Returns:
            Any: Attribute value or wrapped callable managing state.
        """
        attribute = None
        # Check if it's in slots
        if name in getattr(type(self), "__slots__", []):
            return object.__getattribute__(self, name)

        # Check if it's a class-level attribute
        if hasattr(type(self), name):
            return getattr(type(self), name)
        # Set the execution state to active for state tracking
        self.statevars.exc_state = ExecutionState.BEGIN

        # Fetch methods retrieve from the Result object.
        if self.result and hasattr(self.result, name):
            attribute = getattr(self.result, name)
        else:
            if not self.session and not self.connection:
                self.connect()
            # Get attribute from the connection or session
            if self.engine_type == "direct":
                if hasattr(self.connection, name):
                    attribute = getattr(self.connection, name)
            elif hasattr(self.session, name):
                attribute = getattr(self.session, name)

        if not attribute:
            raise AttributeError(f"Attribute {name} not found")
        if callable(attribute):

            @wraps(attribute)
            def wrapper(*args, **kwargs):
                # Update state variables
                try:
                    # Process the args/kwargs and update any based on pre-processors
                    # (e.g. applying textclause to statement strings)
                    if name == "execute":
                        # Process any param handlers and convert str statements to textclause
                        new_args, new_kwargs = self._process_execute_args_kwargs(
                            *args, **kwargs
                        )
                        self.result = result = attribute(*new_args, **new_kwargs)
                    else:
                        self.result = result = attribute(*args, **kwargs)
                    # Add debugging information to the result
                    if isinstance(result, CursorResult) and self.debug:
                        self.update_cursor_meta(result)
                    # Process results if there are result handlers and result is
                    # a valid result type to be processed. Result types can be
                    # added to the result_types list using cls.add_result_type()
                    if (
                        result
                        and self.result_handlers
                        and (
                            isinstance(result, tuple(self.result_types))
                            or (
                                isinstance(result, list)
                                and all(
                                    isinstance(item, tuple(self.result_types))
                                    for item in result
                                )
                            )
                        )
                    ):
                        result = h_(handlers=self.result_handlers, data=result)
                    # Return the result
                    self.statevars.exc_state = ExecutionState.IDLE
                    return result

                except Exception as e:  # pylint: disable=broad-except
                    self.statevars.exc_state = ExecutionState.ERROR
                    # An error handler to capture different errors and apply
                    # handling globally.
                    if self.error_handlers:
                        if isinstance(self.error_handlers, list):
                            for handler in self.error_handlers:
                                handler(e)
                        else:
                            self.error_handlers(e)
                    else:
                        raise e from e

            return wrapper
        self.statevars.exc_state = ExecutionState.IDLE
        return attribute

    def _process_execute_args_kwargs(self, *args, **kwargs):
        """ """
        param_key = "parameters" if self.engine_type == "direct" else "params"

        if args:
            statement = args[0]
            if len(args) > 1:
                params = args[1]
            else:
                params = kwargs.get(param_key, {})
        else:
            statement = kwargs.pop("statement", None)
            params = kwargs.pop(param_key, {})
        # Return the args and let sqlalchemy handle the execution error
        if not statement:
            return args, kwargs

        if isinstance(statement, str) and self.db.apply_textclause:
            statement = text(statement)

        # Process any parameter handlers. This is useful to cleanse or validate
        # any parameters.
        if params and self.parameter_handlers:
            for handler in self.parameter_handlers:
                statement, params = handler(params)

        kwargs["statement"] = statement
        kwargs[param_key] = params

        return (), kwargs

    @property
    def engine(self) -> Engine:
        """
        Create a SQLAlchemy engine. See sqlalchemy.create_engine for more details.

        Renamed from engine to potentially provide some integration with
        flask_sqlalchemy
        """
        if not self.current_engine:
            engine_options = (
                self.db.engine_options.model_dump(
                    exclude_unset=True, exclude_none=True
                )
                if self.db.engine_options
                else {}
            )
            self.current_engine = create_engine(self.url_string, **engine_options)
        return self.current_engine

    def has_connection(self) -> bool:
        """
        Check if there is a connection/session to the database based on engine type.
        """
        if self.engine_type == "direct":
            return True if self.connection else False
        else:
            return True if self.session else False

    def new_session(self) -> ElixirDB:
        """
        Return a new session with the current configuration.

        If there is no session_factory, it means that there is isn't an active
        session either and using this method  will raise an exception.

        In this scenario, just use connect() to create a new session on the current
        instance.

        Returns:
            ElixirDB: New instance with active session.

        Raises:
            InvalidEngineTypeError: If connection mode is DIRECT or no session factory.
            NoSessionFactoryError: If session factory is not set.
        """

        if self.engine_type != "session":
            raise InvalidEngineTypeError(
                "new_session only works with `session` engine_type."
            )
        if not self.session_factory:
            raise NoSessionFactoryError(
                "A session was never created. Use connect() first."
            )

        self.session_factory = self.create_session_factory(
            engine=self.engine, engine_type=self.engine_type
        )
        config = self.config or self.db

        return ElixirDB(
            config=config,
            engine_key=self.engine_key,
            engine_type=self.engine_type,
            session_factory=self.session_factory,
            result_types=self.result_types,
        )

    def connect(self) -> Self:
        """Open, set, and return the connection."""
        if (self.engine_type == "direct" and self.connection) or (
            self.engine_type in ("scoped", "session") and self.session
        ):
            return self

        if self.engine_type == "direct":
            self.connection = self.engine.connect()
        else:
            # Ensure there is a factory
            if self.session_factory is None:
                self.session_factory = self.create_session_factory(
                    engine=self.engine, engine_type=self.engine_type
                )

            if self.engine_type == "session":
                self.session = self.session_factory()
            else:
                # Scoped sessions use factory directly.
                self.session = self.session_factory

        self.statevars.state = ConnectionState.CONNECTED
        return self

    def create_session_factory(
        self,
        engine: Engine | None = None,
        engine_type: EngineType | None = None,
        scopefunc: Any | None = None,
    ) -> sessionmaker[Session] | scoped_session[Session]:
        """
        Helper method to create a session factory.

        See :class:`sqlalchemy.orm.session.sessionmaker` for more details.

        Args:
            engine: SQLAlchemy engine. Default behavior is to check the
                property `engine` and use that.
            engine_type: The engine type to use. `direct` will yield a
                `connection` assigned to self, whereas `session` and `scoped`
                will yield a `session` assigned to self.
            scopefunc: Scope function for scoped sessions.
        Raises:
            InvalidEngineTypeError: If the connection mode is set to `direct`,
            this will raise an error.
        """
        engine = engine or self.engine
        engine_type = engine_type or self.engine_type

        # Extract and prepare session options
        options = (
            self.db.session_options.model_dump(
                exclude_none=True, exclude_unset=True
            )
            if self.db.session_options
            else {}
        )

        if engine_type == "session":
            return sessionmaker(bind=engine, **options)

        # Reference to flask_sqlalchemy implementation. A scopefunc callable
        # can be assigned to the class as well. Used with scoped sessions.
        scopefunc = scopefunc or options.pop("scopefunc", None) or self.scopefunc

        # Cache the scope function if not already set
        self.scopefunc = scopefunc

        if engine_type == "scoped":
            return scoped_session(
                sessionmaker(bind=engine, **options), scopefunc=scopefunc
            )

        raise InvalidEngineTypeError(
            "Cannot create session factory when using direct connection mode."
        )

    def set_engine(self, engine_key: str) -> None:
        """
        Set or change the configuration of the instance in a multi-database
        configuration by key engine_key.

        Args:
            engine_key: The engine_key of the database configuration to use.

        Raises:
            InvalidConfigModel: If attempting to call the method when the
                configuration is not a MultiEngineModel.
            KeyError: If the database configuration with the given engine_key
                does not exist.
        """
        if not self.config:
            raise InvalidElixirConfigError(
                "Cannot set engine when there is not configuration for "
                "multiple engines found. Review the docs for setting up "
                "multiple engines."
            )

        if engine_key not in list(self.config.engines.keys()):
            raise EngineKeyNotFoundError(engine_key=engine_key)

        if self.connection is not None:
            self.close()

        self.engine_key = engine_key
        self.db = self.config.engines[engine_key]
        # Automatically connect new engine
        if self.db.auto_connect and hasattr(self, "connect"):
            self.connect()

    def close(self) -> None:
        """Close the connection to the database and cleanup resources."""
        if not self.connection:
            return
        try:
            if self.engine_type == "scoped" and self.session_factory:
                self.session_factory.remove()
            else:
                self.connection.close()
            self.connection = None
        except Exception as e:
            raise ConnectionError(f"Failed to close connection: {e!s}") from e

    def fetch_results(self, fetch: int | None = None) -> Sequence[RowData]:
        """Fetch results from the result object as mappings."""
        result = self.result
        if not result or not isinstance(self.result, CursorResult):
            raise CursorResultError(
                "The result object does not exist or is not a CursorResult."
            )

        if self.db.result_to_dict:
            if fetch == 0:
                return result.mappings().all()
            return result.mappings().fetchmany(fetch)

        if fetch == 0:
            return result.all()

        return result.fetchmany(fetch)

    def update_cursor_meta(self, result: CursorResult) -> None:
        """Update self.statevars.cursor_meta with metadata from CursorResult."""
        self.statevars.cursor_meta = CursorResultMetadata(
            columns=list(result.keys()),
            rows_affected=result.rowcount,
            last_insert_id=result.lastrowid,
            last_insert_rowid=result.lastrowid,
            last_insert_rowid_str=str(result.lastrowid),
            parameters=getattr(result.context, "compiled_parameters", None),
            sql_query=str(getattr(result.context, "compiled_statement", None)),
            column_descriptions=[
                {"name": col, "type": type_}
                for col, type_ in getattr(result.cursor, "description", []) or []
            ],
            generated_keys=getattr(result, "inserted_primary_key", None),
            is_closed=result.closed,
            execution_context=str(getattr(result, "context", None)),
        )

    def update_orm_meta(self, result: Any) -> None:
        """Update self.statevars.orm_meta with metadata from CursorResult."""

    @classmethod
    def set_scopefunc(cls, scopefunc: Any) -> None:
        """
        Add a scopefunc.

        This adds a scopefunc that the session_factory will use to identify
        the current session.
        """
        cls._scopefunc = scopefunc


class StatementsMixin:
    """
    Mixin class to add support for stored procedures. Will override execute
    method of the base class.
    """

    def procedure(
        self, procedure_name: str, parameters: _CoreAnyExecuteParams | None = None
    ) -> QueryResult:
        """
               Execute a stored procedure with given parameters.

               Args:
                   procedure_name (str): Name of the procedure.
                   parameters (_CoreAnyExecuteParams
        | None): Optional procedure
                       parameters.

               Returns:
                   QueryResult: Dict of affected rows or procedure results.

               Raises:
                   ValueError: If executed on SQLite.
        """

        dialect = self.db.dialect

        if dialect == "sqlite":
            raise ValueError(
                "SQLite does not support stored procedures or functions."
            )

        # Apply a schema prefix if enabled
        if (
            not self.statevars.schema_applied
            and self.db.statements.prefix_procedures
        ):
            procedure_name = self.add_schema_prefix(procedure_name, "procedure")
            self.statevars.schema_applied = True

        # Build the parameter string
        proc_parameters = self._build_parameters(parameters) if parameters else ""
        # Build the statement
        sql_statement = self._build_proc_stmt(
            procedure_name=procedure_name,
            parameters_str=proc_parameters,
            dialect=self.db.dialect,
        )

        return self.execute(sql_statement, parameters)

    def _execute_query(
        self,
        statement: Executable | str,
        parameters: _CoreAnyExecuteParams | None = None,
        fetch: int | None = None,
    ) -> Sequence[RowMapping] | Sequence[Any]:
        """Execute the actual statement and handle result fetching"""
        if parameters:
            result = self.result = self.connection.execute(statement, parameters)
        else:
            result = self.result = self.connection.execute(statement)
        if not result.returns_rows:
            return result

        self.statevars.columns = result.keys()

        if fetch is not None:
            result = self.fetch_results(result, fetch)

        return result

    def _build_proc_stmt(
        self,
        procedure_name: str,
        parameters_str: str,
        dialect: str,
    ) -> str:
        """Build SQL template for stored procedure based on dialect"""
        match self.db.dialect:
            case "postgres" | "mysql":
                return f"CALL {procedure_name}{parameters_str}"
            case "mssql":
                return f"EXEC {procedure_name} {parameters_str}"
            case "oracle":
                return f"BEGIN {procedure_name}{parameters_str}; END;"
            case _:
                raise ValueError(f"Unsupported dialect: {dialect}")

    def _build_parameters(self, parameters: _CoreAnyExecuteParams) -> str:
        """Build SQL template for stored procedure based on dialect"""

        match self.db.dialect:
            case "mssql" if parameters:
                return ", ".join(f"@{key} = :{key}" for key in parameters)
            case _:
                return (
                    f"({', '.join(f':{key}' for key in parameters)})"
                    if parameters
                    else ""
                )

    def add_schema_prefix(
        self,
        statement: str,
        operation_type: Literal["statement", "procedure"] = "procedure",
    ) -> str:
        """Add the schema to stored procedure/function names."""
        if not self.db.statements.schema_name:
            return statement
        if operation_type == "procedure":
            if "." in statement:
                warnings.warn(
                    "Prefix requires a query, stored procedure, or function name.",
                    stacklevel=2,
                )
            return f"{self.db.statements.schema_name}.{statement}"

        return apply_schema_to_statement(
            statement, self.db.config.schema_name, self.db.dialect
        )


class ElixirDBStatements(StatementsMixin, ElixirDB):
    """Statement enabled connection class."""


def create_db(
    config: DatabaseEngineConfig | None,
    engine_key: str | None = None,
    enable_statements: bool = False,
    engine_type: EngineType = "direct",
    **kwargs: Any,
) -> ElixirDB:
    """
    Create a db connection instance.

    Args:
        config (DbConfigDict | Config | None): Database configuration.
        engine_key (str | None): Connection engine_key.
        enable_statements (bool): If True, enable statement mixin.
        **kwargs: Additional keyword arguments for connection initialization.
            These can be a handler_dict, or other attributes for the base
            connection config class.

    Returns:
        ElixirDB: An instance of ElixirDB or a subclass
        with statement execution capabilities.
    """
    if enable_statements:
        return ElixirDBStatements(
            config=config,
            engine_key=engine_key,
            engine_type=engine_type,
            **kwargs,
        )

    return ElixirDB(
        config=config, engine_key=engine_key, engine_type=engine_type, **kwargs
    )



"""Options for configuring SQLAlchemy connections."""

# ruff:: noqa: E501
# flake8: noqa: E501
# pylint: disable=C0301
from __future__ import annotations

from typing import Any
from typing import Callable
from pydantic import Field
from sqlalchemy.engine import Engine
from typing_extensions import Annotated
from elixirdb.models.model import StrictModel
from elixirdb.models.schemas import CallableSchema
from elixirdb.models.schemas import EngineSchema


Engine_ = Annotated[Engine, EngineSchema]
Callable_ = Annotated[Callable, CallableSchema]


class SessionOptions(StrictModel):
    """
    Options for configuring SQLAlchemy sessions.
    See :class:`sqlalchemy.orm.session.Session` for more details.
    """

    autocommit: bool = Field(
        False,
        description=(
            "When enabled, the session will automatically commit changes after "
            "each operation. This can be useful for simple operations but may not be "
            "appropriate for complex scenarios."
        ),
    )
    autoflush: bool = Field(
        True,
        description=(
            "When enabled, ensures that the session automatically calls "
            "`Session.flush()` before executing any query. This eliminates the "
            "need to manually flush changes, ensuring queries always operate on "
            "the latest state."
        ),
    )
    expire_on_commit: bool = Field(
        True,
        description=(
            "When set to True, all objects in the session are automatically expired "
            "after `commit()`, ensuring their state is reloaded from the database upon "
            "next access. This guarantees consistency with the database's current state."
        ),
    )
    twophase: bool = Field(
        False,
        description=(
            "When enabled, transactions use a two-phase commit protocol, allowing "
            "preparation of transactions across multiple databases before a final commit. "
            "This is useful for distributed or multi-database transactions."
        ),
    )
    bind: Engine_ | None = Field(
        None,
        description=(
            "The `bind` parameter associates a specific `Engine` or `Connection` "
            "with the session, ensuring all SQL operations are executed via this connection."
        ),
    )

    binds: dict[str, Engine_] | None = Field(
        None,
        description=(
            "A dictionary mapping specific entities (e.g., tables, mapped classes) "
            "to individual `Engine` objects for per-entity or multi-engine setups."
        ),
    )


class EngineOptions(StrictModel):
    """
    SqlAlchemy-specific options for database configurations.

    See :class:`sqlalchemy.engine.Engine` for more information or visit
    https://docs.sqlalchemy.org/en/20/core/engines.html#sqlalchemy.create_engine
    """

    connect_args: dict[str, Any] | None = Field(
        None,
        description=(
            "Dictionary of keyword arguments passed directly to the DBAPI's "
            "connect() method. Allows customization of the database "
            "connection process. See SQLAlchemy docs on 'Custom DBAPI "
            "connect() arguments / on-connect routines'."
        ),
    )

    creator: Callable_ | None = Field(
        None,
        description=(
            "Callable that returns a DBAPI connection. Used by the connection "
            "pool to create new connections. This bypasses connection "
            "parameters specified in the URL. DialectEvents.do_connect() "
            "offers more flexibility for customization."
        ),
    )

    echo: bool | str = Field(
        False,
        description=(
            "If True, logs all SQL statements and their parameters. If set "
            "to 'debug', result rows are also printed. Uses Python logging."
        ),
    )

    echo_pool: bool | str = Field(
        False,
        description=(
            "If True, logs pool activity like invalidation and recycling. If "
            "set to 'debug', checkouts and checkins are logged. Uses Python "
            "logging for more control."
        ),
    )

    enable_from_linting: bool = Field(
        True,
        description=(
            "If True, emits warnings for SELECTs with unlinked FROM elements "
            "that may cause Cartesian products."
        ),
    )

    execution_options: ExecutionOptions | None = Field(
        None,
        description=(
            "Dict of execution options applied to all connections. See "
            "Connection.execution_options() for details."
        ),
    )

    future: bool = Field(
        True,
        description=(
            "Use the 2.0-style Engine and Connection API. Default and only "
            "supported mode in SQLAlchemy 2.x."
        ),
    )

    hide_parameters: bool | None = Field(
        None,
        description=(
            "If True, hides SQL parameters in INFO logs and StatementError "
            "strings. Useful for logging sensitive data."
        ),
    )

    insertmanyvalues_page_size: int | None = Field(
        None,
        description=(
            "Rows per INSERT statement in 'insertmanyvalues' mode. Defaults "
            "to 1000 but varies by dialect."
        ),
    )

    isolation_level: str | None = Field(
        None,
        description=(
            "Transaction isolation level for new connections. Overrides URL "
            "defaults. See SQLAlchemy docs on 'Setting Transaction "
            "Isolation Levels'."
        ),
    )

    json_deserializer: Callable_ | None = Field(
        None,
        description=(
            "Callable to deserialize JSON strings into Python objects. "
            "Defaults to json.loads."
        ),
    )

    json_serializer: Callable_ | None = Field(
        None,
        description=(
            "Callable to serialize Python objects into JSON strings. Defaults "
            "to json.dumps."
        ),
    )

    label_length: int | None = Field(
        None,
        description=(
            "Max length of generated column labels. Defaults to the "
            "dialect's max identifier length."
        ),
        gt=0,
    )

    logging_name: str | None = Field(
        None,
        description=(
            "Identifier used in SQLAlchemy engine log messages. Defaults to "
            "a hexstring of the object's id."
        ),
    )

    max_identifier_length: int | None = Field(
        None,
        description=(
            "Overrides dialect max identifier length. Defaults to dialect settings."
        ),
    )

    max_overflow: int = Field(
        10,
        description=(
            "Max connections allowed in pool overflow. Only for QueuePool."
        ),
    )

    paramstyle: str | None = Field(
        None,
        description=(
            "DBAPI parameter style (e.g., 'named', 'qmark'). Defaults to the "
            "DBAPI's default."
        ),
    )

    pool: Any | None = Field(
        None,
        description=(
            "Pre-constructed Pool instance for connection pooling. Overrides "
            "URL parameters."
        ),
    )

    poolclass: Any | None = Field(
        None,
        description=(
            "Pool subclass used for connection pooling. Specifies type but "
            "does not instantiate it."
        ),
    )

    pool_logging_name: str | None = Field(
        None,
        description=(
            "Identifier for pool log messages. Defaults to object's id hexstring."
        ),
    )

    pool_pre_ping: bool | None = Field(
        None,
        description=(
            "If True, enables connection liveness checks before checkout "
            "from the pool."
        ),
    )

    pool_size: int = Field(
        5,
        description=(
            "Number of connections in the pool. 0 disables pooling "
            "(use NullPool instead)."
        ),
    )

    pool_recycle: int = Field(
        -1,
        description=(
            "Seconds before a connection is recycled. -1 disables timeout."
        ),
    )

    pool_reset_on_return: str | None = Field(
        "rollback",
        description=(
            "Reset strategy on connection return: 'rollback', 'commit', or "
            "None. See SQLAlchemy docs on 'Reset On Return'."
        ),
    )

    pool_timeout: float = Field(
        30,
        description=(
            "Seconds to wait for a connection from the pool before timing out."
        ),
    )

    pool_use_lifo: bool = Field(
        False,
        description=(
            "If True, uses LIFO for connection retrieval in QueuePool "
            "(default is FIFO)."
        ),
    )

    plugins: list[str] | None = Field(
        None,
        description=("List of plugin names to load. See CreateEnginePlugin docs."),
    )

    query_cache_size: int = Field(
        500,
        description=(
            "Size of SQL string cache. 0 disables caching. Cache pruned "
            "when 1.5x size."
        ),
    )

    use_insertmanyvalues: bool = Field(
        True,
        description=(
            "If True, enables 'insertmanyvalues' mode for bulk inserts with "
            "RETURNING clauses."
        ),
    )


class ExecutionOptions(StrictModel):
    """
    Execution options for SQLAlchemy connections.

    See :class:`sqlalchemy.engine.Connection.execution_options` or visit
    https://docs.sqlalchemy.org/en/20/core/connections.html# for more details.

    """

    compiled_cache: dict[Any, Any] | None = Field(
        None,
        description=(
            "Dictionary for caching compiled SQL statements. This can "
            "improve performance by reusing parsed query plans instead of "
            "recompiling them each time a query is executed."
        ),
    )
    logging_token: str | None = Field(
        None,
        description=(
            "A token included in log messages for debugging concurrent "
            "connection scenarios. Useful for tracking specific database "
            "connections in a multi-threaded or multi-process environment."
        ),
    )

    isolation_level: str | None = Field(
        None,
        description=(
            "Specifies the transaction isolation level for this connection. "
            "Controls how transactions interact with each other. Common "
            "values include 'SERIALIZABLE', 'REPEATABLE READ', "
            "'READ COMMITTED', 'READ UNCOMMITTED', and 'AUTOCOMMIT'."
        ),
    )
    no_parameters: bool | None = Field(
        None,
        description=(
            "If True, skips parameter substitution when no parameters are "
            "provided. Helps prevent errors with certain database drivers "
            "that treat statements differently based on parameter presence."
        ),
    )
    stream_results: bool | None = Field(
        None,
        description=(
            "Enables streaming of result sets instead of pre-buffering "
            "them in memory. Useful for handling large query results "
            "efficiently by fetching rows in batches."
        ),
    )
    max_row_buffer: int | None = Field(
        None,
        description=(
            "Defines the maximum buffer size for streaming results. "
            "Larger values reduce query round-trips but consume more memory. "
            "Defaults to 1000 rows."
        ),
    )
    yield_per: int | None = Field(
        None,
        description=(
            "Specifies the number of rows to fetch per batch when streaming "
            "results. Optimizes memory usage and improves performance "
            "for large result sets."
        ),
    )
    insertmanyvalues_page_size: int | None = Field(
        None,
        description=(
            "Determines how many rows are batched into an INSERT statement "
            "when using 'insertmanyvalues' mode. Defaults to 1000 but "
            "varies based on database support."
        ),
    )
    schema_translate_map: dict[str, str] | None = Field(
        None,
        description=(
            "A mapping of schema names for automatic translation during "
            "query compilation. Useful for working across multiple schemas "
            "or database environments."
        ),
    )
    preserve_rowcount: bool | None = Field(
        None,
        description=(
            "If True, preserves row count for all statement types, "
            "including SELECT and INSERT, in addition to the default "
            "behavior of tracking row counts for UPDATE and DELETE."
        ),
    )



"Model Validation Errors"

from __future__ import annotations


REF_URL = "https://github.com/hotnsoursoup/quik-db/blob/master/README.md%s"


class ModelConfigError(Exception):
    """
    Custom exception class for handling validation errors.
    """


def print_and_raise_validation_errors(validation_errors, raise_error=True):
    """
    Prints validation errors extracted from a ValidationError object
    with linked information.

    """

    errors = validation_errors.errors()  # Extract errors from ValidationError
    config_errors = []
    engine_errors = {}
    msg = ""

    for error in errors:
        # Extract location and separate root config errors and engine model errors.
        loc = error.get("loc", [])
        if loc and loc[0] == "engines" and len(loc) > 1:
            name = loc[1]
            if name not in engine_errors:
                engine_errors[name] = [error]
            else:
                engine_errors[name].append(error)
        else:
            config_errors.append(error)

    if config_errors:
        msg = "Configuration Errors:\n"

        for e in config_errors:
            msg += add_model_error_to_msg(e)

    if engine_errors:
        for k, v in engine_errors.items():
            msg += f"Engine Errors: {k}\n"
            for engine_error in v:
                msg += add_model_error_to_msg(engine_error)
    if raise_error:
        msg = "\n" + msg
        raise ModelConfigError(msg) from None
    else:
        print(msg)
        return msg


def add_model_error_to_msg(error: dict) -> str:
    """
    Add an error message to a given error dictionary.
    """
    msg = ""
    location = None
    loc = error.get("loc", [])
    error_type = error.get("type", None)
    input = error.get("input", None)

    if loc and loc[0] == "engines":
        location = ".".join(str(loc) for loc in loc[2:])
    else:
        location = ".".join(str(loc) for loc in loc)

    msg += f"   - Type: {error_type} \n" if error_type else ""
    msg += f"     Location: {location}\n" if location else ""
    msg += f"     Message: {error.get('msg', 'Unknown error message')}\n"
    if error_type and error_type != "missing":
        msg += f"     Input: {input}\n" if input else ""

    # Add additional context if present
    for key, value in error.get("ctx", {}).items():
        if key != "error":
            msg += f"     {key.capitalize()}: {value}\n"
    return msg


def get_error_msg(error: dict) -> str:
    """
    Extract and return an error message from a given error dictionary.

    Args:
        error (dict): A dictionary containing error information.

    Returns:
        str: The extracted error message. If the "ctx" dictionary
            contains an "error" key, its value is returned. Otherwise,
            the value of the "msg" key in the error dictionary is
            returned. If neither is present, an empty string is returned.

    """
    ctx = error.get("ctx", {})
    if ctx and ctx.get("error"):
        return ctx.pop("error")
    return error.get("msg", "")


class ElixirConfigError(Exception):
    """
    Base exception class for ElixirDB configuration errors.
    """


class InvalidElixirConfigError(ElixirConfigError):
    """Exception raised when an invalid configuration is provided."""


class ElixirFileNotFoundError(ElixirConfigError):
    """Exception raised when a specified configuration name is not found."""


class EngineKeyNotFoundError(ElixirConfigError):
    """Exception raised when a specified configuration name is not found."""

    def __init__(self, engine_key: str, message: str | None = None):
        if not message:
            message = f"Engine key '{engine_key}' not found in the configuration."
        message = message.format(engine_key=engine_key)

        super().__init__(message)


class EngineKeyNotDefinedError(ElixirConfigError):
    """
    Exception raised when a configuration could not be identified because
    an 'engine_key' was not provided or a 'default' configuration could not
    be found.

    :class:'elixirdb.models.manager.Config'
    """


class InvalidEngineTypeError(Exception):
    """
    Exception raised when attempting to create a session factory in
    direct connection mode.
    """


class CursorResultError(Exception):
    """
    Exception when attempting to access a cursor result.
    """


class NoSessionFactoryError(Exception):
    """
    Exception raised when attempting to access a session factory.
    """



"Type definitions"

# isort: off
# ruff: noqa: E266, E501
# flake8: noqa: E266, E501
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal, Mapping, Union
from typing import Callable
from typing import NotRequired
from typing import Protocol
from typing import Required
from typing import TypeAlias
from typing import TypedDict
from typing import TypeVar
from pydantic import BaseModel
from sqlalchemy.orm import Session
from sqlalchemy import RowMapping, Row

from elixirdb.models.engine import EngineModel
from elixirdb.enums import Dialect

if TYPE_CHECKING:
    from elixirdb.models.manager import EngineManager
    from sqlalchemy.sql.base import Executable
    from sqlalchemy.engine import Engine
    from sqlalchemy.engine.interfaces import _CoreAnyExecuteParams

M = TypeVar("M", bound=BaseModel)
_S = TypeVar("_S", bound=Session)

# Handler related types
# Dialect = Literal["mysql", "postgres", "sqlite", "oracle", "mssql", "mariadb"]

HandlerCallable = Callable[..., Any]
HandlerNames: TypeAlias = Literal[
    "parameter_handlers", "result_handlers", "error_handlers"
]

ProcessorFunc: TypeAlias = Callable[..., Any]
HandlerSequence: TypeAlias = Sequence[ProcessorFunc]
HandlerMapping: TypeAlias = Mapping[str, HandlerSequence]
# Query related types
SingleParam = Mapping[str, Any]
MultiParam = Sequence[SingleParam]
QueryResult: TypeAlias = (
    list[dict[str, Any]]  # List of dictionaries (e.g., rows with column names)
    | dict[str, Any]  # A single dictionary (e.g., a single row result)
    | list[tuple[Any, ...]]  # List of tuples (e.g., rows without column names)
    | None  # No result or empty result
)

# Database related types
SQLStatement: TypeAlias = str
TableName: TypeAlias = str
SchemaName: TypeAlias = str
DialectName: TypeAlias = str
DatabaseKey: TypeAlias = str
DriverName: TypeAlias = str  # e.g. "postgresql+psycopg2"
DriverMapping: TypeAlias = Mapping[Dialect, DriverName]


EngineType: TypeAlias = Literal["direct", "session", "scoped"]
SessionType: TypeAlias = Session | Callable[[], Session]
RowData: TypeAlias = Row[Any] | RowMapping | list[Row] | Any


# Typed Dicts
class BaseEngineDict(TypedDict, total=False):
    """
    Base TypedDict for shared optional fields in database configurations.
    """

    cursor_result_to_dict: bool
    apply_textclause: bool
    description: str
    default: bool
    engine_options: EngineOptionsDict | Mapping[str, Any]
    session_options: SessionOptionsDict | Mapping[str, Any]


class SessionOptionsDict(TypedDict, total=False):
    """
    SqlAlchemy-specific options for sessionmaker.

    See :class:`sqlalchemy.orm.session.sessionmaker` for more information.

    """

    # Whether to automatically flush pending changes to the database.
    autoflush: bool
    # Whether to expire all ORM objects after commit so that their state
    # is refreshed on next access.
    expire_on_commit: bool
    # Whether to use two-phase commit (requires backend support).
    twophase: bool
    # A default Engine used for this session (if any).
    bind: Engine | None
    # A mapping of bind keys to Engine objects for per-entity or multi-engine setups.
    binds: dict[str, Engine] | None


class EngineOptionsDict(TypedDict):
    """
    SqlAlchemy-specific options for database configurations.

    See :class:`sqlalchemy.engine.Engine` for more information.
    """

    # Database URL and connection
    url: str | None

    # Number of connections in the connection pool
    pool_size: int | None

    # Maximum number of connections to allow beyond the pool size
    max_overflow: int | None

    # Timeout in seconds for getting a connection from the pool
    pool_timeout: int | None

    # Number of seconds before recycling a connection
    pool_recycle: int | None

    # Test connections for liveness before using
    pool_pre_ping: bool | None

    # Log all SQL statements
    echo: bool | None

    # Enable SQLAlchemy 2.0-style behavior (for forward compatibility)
    future: bool | None

    # Connection Arguments
    # Pass arguments to the DBAPI
    connect_args: dict[str, str | int | bool] | None

    # Options for statement execution
    execution_options: dict[str, str | int | bool] | None

    # Thread/Process Management
    # Custom connection pool class
    poolclass: Callable | None

    # Custom connection function
    creator: Callable | None

    # Transaction isolation level (e.g., "READ_COMMITTED")
    isolation_level: str | None

    # Statement Cache
    # Enable/disable native Unicode handling
    use_native_unicode: bool | None

    # Number of prepared statements to cache
    statement_cache_size: int | None

    # Dialect-specific options (common examples)
    # Parameter style, e.g., "qmark", "numeric"
    paramstyle: str | None

    # Case sensitivity in column lookups
    case_sensitive: bool | None

    # Others
    # Log connection pool checkouts/checkins
    echo_pool: bool | None

    # Disable server-side cursors for certain DBs
    disable_server_side_cursors: bool | None


class UrlParams(TypedDict):
    """
    Connection parameters for database connections.

    See :class:`sqlalchemy.engine.URL` for more information.
    """

    drivername: NotRequired[str]
    host: Required[str]
    port: NotRequired[str | int]
    username: NotRequired[str]
    password: NotRequired[str]
    database: NotRequired[str]
    query: NotRequired[Mapping[str, Any]]


class EngineParams(TypedDict):
    """
    TypedDict for configurations requiring url_params and dialect.
    """

    url_params: Required[UrlParams]
    dialect: Required[Dialect]


class EngineUrl(TypedDict):
    """
    TypedDict for configurations requiring url_params and dialect.
    """

    url: Required[str]  # Database connection url string
    dialect: Required[Dialect]


class OptUrl(EngineUrl, BaseEngineDict, total=False):
    """
    TypedDict for default config that uses UrlConfig where the attributes
    are not required.
    """


class OptParam(EngineParams, BaseEngineDict, total=False):
    """
    TypedDict for default config that uses UrlParams where the
    attributes are not required.

    """


# The default configuration applied to all database connections.
DefaultEngineConfig: TypeAlias = OptUrl | OptParam
# Engine Configurations
EngineConfig: TypeAlias = EngineUrl | EngineParams
EngineConfigDict: TypeAlias = Mapping[str, EngineConfig | Mapping[str, Any]]


class ConfigDict(TypedDict, total=False):
    """
    Set the configuration for the application. You can provide a default
    """

    # A database driver mapping dictionary
    driver_mapping: DriverMapping

    # Controls whether settings here override all other settings.
    # Default behavior is for settings in specific databases to override
    # the defaults set in defaults.
    global_override: bool

    # Defaults that are used by all database connections. Any setting
    # in the database config itself will override the defaults set here.
    # The default configuration does not have to validate to a EngineModel.
    # It can be a partial that is inherited by other configs to be a complete
    # configuration.
    defaults: DefaultEngineConfig | Mapping[str, Any] | None


class AppConfigDict(TypedDict):
    """Database configuration dictionary"""

    app: NotRequired[ConfigDict]
    # A dictionary of database configurations, with keys as database names
    # Added dict[str, Any] to allow for more lax type checking
    engines: EngineConfigDict | dict[str, Any]


ConfigModel: TypeAlias = Union["EngineManager", EngineModel]

# The configuration passed into the connection classes
DatabaseEngineConfig: TypeAlias = (
    EngineConfig | AppConfigDict | ConfigModel | Mapping[str, Any]
)


# Protocol definitions
class ExecutionProtocol(Protocol):
    """
        This protocol specifies the methods that must be implemented by any
        class that wants to serve as a database executor.


        Example:
            >>> class CustomExecutionHandler(ExecutionProtocol):
            ...     def __call__(
            ...         self,
            ...         statement: str,
            ...         params: _CoreAnyExecuteParams
    ,
            ...         fetch: int | None
            ...     ) -> QueryResult:
            ...         # Your custom execution logic here
            ...         return result
    """

    def __call__(
        self, statement: str | Executable, params: _CoreAnyExecuteParams
    ) -> QueryResult: ...


class ErrorHandlerProtocol(Protocol):
    """
    Defines the interface for an error handler that can intercept and
    process exceptions. Classes implementing this protocol may also
    intercept attribute access with __getattr__ if desired.

    Example:
        >>> class CustomErrorHandler(ErrorHandlerProtocol):
        ...     def __init__(self, logger: logging.Logger):
        ...         self.logger = logger
        ...
        ...     def __call__(self, error: Exception) -> Any:
        ...         self.logger.error(f"An error occurred: {error}")
        ...         return "Handled result"

    """

    def __call__(self, error: Exception, config: EngineConfig | None = None) -> Any:
        """
        Handle an exception and return an appropriate response or action.

        Args:
            error (Exception): The exception to handle.

        Returns:
            Any: The result of handling the exception. Could be a default
                 value, a re-raised or wrapped exception, or any other
                 custom outcome.
        """

    def __getattr__(self, attr: str) -> Any:
        """
        Optionally intercept attribute/method access for additional error
        handling behavior, if needed.

        This is left open-ended to allow dynamic behavior. For example,
        you might dynamically generate methods that always handle errors
        internally.

        Args:
            attr (str): The attribute name being accessed.

        Returns:
            Any: The attribute or a callable that handles errors.
        """


class ResultHandlerCallable(Protocol):
    """
    Protocol for a result handler function. This function is responsible
    for processing the output of a database operation and returning the
    result for any result_types in ElixirDB.result_types.

    Example:
        >>> def filter_results(result: Any) -> Any:
        ...     # Your custom result processing logic here
        ...     return result
    """

    def __call__(
        self,
        result: Any,
        **kwargs: Any,
    ) -> Any: ...


class ParamHandlerCallable(Protocol):
    """
    Protocol to do parameter pre-processing before a query is executed
    (e.g. for logging, cleansing,)
    """

    def __call__(
        self,
        parameters: _CoreAnyExecuteParams,
        config: DatabaseEngineConfig | None = None,
        **kwargs: Any,
    ) -> Any: ...



"""
Validation Module for database models.

This module provides a validator function for validating any db_models
against a given database configuration. It uses Pydantic for validation and

"""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Type
from pydantic import Field
from pydantic import ValidationError
from pydantic_core import InitErrorDetails
from typing_extensions import Annotated


if TYPE_CHECKING:
    from elixirdb.types import M


PositiveInt = Annotated[int, Field(gt=0)]
ConnectArgsType = dict[str, Any]
ExecutionOptionsType = dict[str, Any]
JsonDeserializerType = Callable[[str], Any]
JsonSerializerType = Callable[[Any], str]
PluginListType = list[str]


def model_validator(db_config: dict[str, Any], models: list[Type[M]]) -> M | None:
    """
    Validates the given database configuration against a list of models.

    Args:
        config: The config to validate against the models.
        models: A list of model classes to validate against.

    Returns:
        The validated model instance or None if validation fails.
    """
    errors = []

    def validate_model(model_class: Type[M]) -> M | None:
        try:
            return model_class(**db_config)
        except ValidationError as e:
            errors.append(e.errors())
            return None

    for model in models:
        validated_model = validate_model(model)
        if validated_model:
            return validated_model
    if errors:
        raise ValidationError([InitErrorDetails(err) for err in errors])



"""
Application Configuration
"""

# ruff: noqa: TC001
from __future__ import annotations

from typing import Any
from typing import ClassVar
from typing import Mapping
from typing import Self
from pydantic import Field
from pydantic import model_validator

# from elixirdb.dictionaries.resources import drivers
from elixirdb.models.engine import EngineModel
from elixirdb.models.engine import driver_map
from elixirdb.models.model import StrictModel
from elixirdb.types import DatabaseKey
from elixirdb.types import DriverMapping


class EngineManager(StrictModel):
    """
    Configuration Manager for SqlAlchemy Engines.

    Sample Engine dictionary.

    engine = {
        "dialect": "postgres",
        "url": "postgresql+psycopg2://test_user:StrongPassword!123@localhost:5432/ellixirdb"
        "engine_options": {
            "pool_size": 10,
            "pool_recycle": 3600,
            "pool_pre_ping": True,
        },
        "session_options": {
            "autocommit": True,
            "autoflush":
        },
        "execution_options": {},
        "default": True
    }

    }
    """

    default_engine_key: ClassVar[DatabaseKey | None] = None

    defaults: Mapping[str, Any] | None = Field(
        None,
        description=(
            "A mapping of EngineModel values. All fields become optional "
            "and these values are inherited by all engines. Values within "
            "each engine will take precedence over the defaults."
        ),
    )

    engines: Mapping[DatabaseKey, EngineModel] = Field(
        description=(
            "A dictionary of engine configurations. The key is the "
            "dialect and the value is an EngineModel."
        )
    )

    driver_mapping: DriverMapping = Field(
        driver_map,
        description=(
            "A dictionary that maps a database dialect to a "
            "'driver' or 'dialect+driver'. "
        ),  # pyright: ignore[reportAssignmentType]
    )
    global_override: bool = Field(
        False,
        description=(
            "Controls whether settings here override all other settings. "
            "Default behavior is for settings in specific databases to override "
            "the defaults set in defaults. "
        ),
    )

    debug: bool = Field(
        False,
        description=("Controls debug mode for all engines"),
    )

    @model_validator(mode="before")
    @classmethod
    def update_configs_with_default_values(
        cls,
        values: dict[str, Any],
    ) -> dict[str, Any]:
        """
        Set the default values that all engines will use as a base. Configurations
        in the engines dictionary will override the defaults.
        """
        engines = values.get("engines", None)
        defaults = values.get("defaults", None)
        msg = "The `%s` key must be a dictionary or mapping."
        if engines and not isinstance(engines, dict):
            raise ValueError(msg % "engines")
        if defaults and not isinstance(defaults, dict):
            raise ValueError(msg % "defaults")

        if engines and defaults:
            if defaults:
                for engine_name, engine in values["engines"].items():
                    # Merge defaults and engine, with engine overriding defaults
                    merged_engine = defaults.copy()
                    merged_engine.update(engine)
                    values["engines"][engine_name] = merged_engine
        return values

    @model_validator(mode="after")
    def ensure_only_one_default(self) -> Self:
        """Ensure only only one database is marked as default"""
        if len(self.engines) < 1:
            return self

        defaults = sum(1 for db in self.engines.values() if db.default)

        if defaults > 1:
            raise ValueError(
                "Only one database can be set as the default. "
                "Please remove extra 'default' flags and ensure only one is set."
            )
        return self

    @model_validator(mode="after")
    def get_and_set_default_engine_key(self) -> Self:
        """
        Set the default engine key if one is set.
        """
        if not self.engines:
            return self
        for engine_key, engine in self.engines.items():
            if engine.default:
                self.set_default_engine_key(engine_key)
        return self

    @classmethod
    def set_default_engine_key(cls, engine_key: DatabaseKey) -> None:
        """
        Set the default engine key for the class.
        """
        cls.default_engine_key = engine_key



"""
Database base configuration classes.

"""

# ruff: noqa: D102
# flake8: noqa: E501
# pylint: disable=C0301
# pyright: reportAttributeAccessIssue=false, reportArgumentType=false, reportUnknownVariableType=false
from __future__ import annotations

import warnings
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from typing import Any
from typing import ClassVar
from pydantic import ValidationError
from sqlalchemy import URL
from sqlalchemy import Connection
from sqlalchemy import CursorResult
from sqlalchemy import Engine
from sqlalchemy import Result
from sqlalchemy import ScalarResult
from sqlalchemy.orm.decl_api import declarative_base
from elixirdb.enums import ConnectionState
from elixirdb.enums import ExecutionState
from elixirdb.exc import ElixirFileNotFoundError
from elixirdb.exc import EngineKeyNotDefinedError
from elixirdb.exc import EngineKeyNotFoundError
from elixirdb.exc import print_and_raise_validation_errors
from elixirdb.models.engine import EngineModel
from elixirdb.models.manager import EngineManager
from elixirdb.utils.files import load_config


if TYPE_CHECKING:
    from sqlalchemy.orm import scoped_session
    from sqlalchemy.orm import sessionmaker
    from sqlalchemy.orm.session import Session
    from elixirdb.types import DatabaseEngineConfig
    from elixirdb.types import EngineType
    from elixirdb.types import ErrorHandlerProtocol
    from elixirdb.types import ExecutionProtocol
    from elixirdb.types import HandlerMapping
    from elixirdb.types import ParamHandlerCallable
    from elixirdb.types import ResultHandlerCallable
    from elixirdb.types import T


@dataclass(slots=True)
class ORMResultMetadata:
    """Metadata for ORM result objects."""

    model_name: str | None = None
    total_rows: int | None = None
    query_time: float | None = None
    fields: list[str] = field(default_factory=list)
    relationships: list[str] = field(default_factory=list)
    rows: list[Any] = field(default_factory=list)


@dataclass(slots=True)
class CursorResultMetadata:
    """ "Metadata for CursorResult objects"""

    columns: list[str] | None = None
    rows_affected: int | None = None
    last_insert_id: Any | None = None
    last_insert_rowid: Any | None = None
    last_insert_rowid_str: str | None = None
    pre_format_parameters: dict[str, Any] | list[Any] | None = None
    parameters: dict[str, Any] | list[Any] | None = None
    pre_format_sql_query: str | None = None
    sql_query: str | None = None
    column_descriptions: list[dict[str, Any]] = field(default_factory=list)
    generated_keys: list[Any] | None = None
    is_closed: bool | None = None
    execution_context: str | None = None


@dataclass(slots=True)
class StateVars(CursorResultMetadata):
    """State variables for tracking database execution state"""

    state: ConnectionState = ConnectionState.DISCONNECTED
    exc_state: ExecutionState = ExecutionState.IDLE
    connection_type: str = ""
    schema_applied: bool = False
    fetch_size: int | None = None
    page: int = 0
    limit: int = 0
    offset: int = 0
    rowcount: int = 0
    results: list = field(default_factory=list)
    cursor_meta: CursorResultMetadata = field(default_factory=CursorResultMetadata)
    orm_meta: ORMResultMetadata = field(default_factory=ORMResultMetadata)


@dataclass(slots=True)
class ConnectionConfig:
    """Boilerplate database configuration for SqlAlchemy"""

    # The model representation of the database configuration. This
    # will be generated based on the db_config.
    db: EngineModel

    # The global app or database settings.
    config: ClassVar[EngineManager | None] = None

    # Enable debug mode for more verbose state variable tracking
    debug: bool = False

    # Bypass validation. For internal use.
    _bypass: bool = False

    # The key used to identify the database configuration that is loaded.
    engine_key: str = ""

    # Result object returned from the database execution. This can be a
    # SqlAlchemy Result object, a declarative base class, SQLModel, etc.
    result: CursorResult | T | list[T] | None = field(default=None)

    # The current sqlalchemy engine being used by the subclass. Created
    # with :class:`sqlalchemy.create_engine`
    current_engine: Engine | None = None

    # The result types that result_handlers will process.
    # e.g. [Result, SqlAlchemyResult, SqlAlchemyRow, SqlAlchemyRowProxy]
    # All handlers in result_handlers will process the output of any
    # type defined in the result_types in addition to the SqlAlchemy
    # declarative base class.

    # Some types that may be useful for result_types:
    # Result, Row, RowMapping, ScarlarResult, MappingResult, ChunkedIteratorResult

    # This is not comprehensive as it does not include all ORM type results such
    # as Query, QueryResult, etc.
    result_types: list[Any] = field(default_factory=list)

    # The database connection created using a `direct` engine_type.
    connection: Connection | None = field(default=None)

    # The session created using a `session` or `scoiped_session` engine_type.
    session: Session | scoped_session[Session] | None = field(default=None)

    # The connection mode - direct, session, or scoped_session.
    engine_type: EngineType = "direct"

    session_options: dict[str, Any] | None = None
    # A session factory if the connection mode is session or scoped_session.
    session_factory: sessionmaker[Session] | scoped_session[Session] | None = field(
        default=None
    )

    # State tracking variables. Mainly used for debugging
    statevars: StateVars = field(default_factory=StateVars)

    # A custom handler to control the execution process for statements
    # such as raw SQL, Stored Procedures, etc.
    execution_handler: ExecutionProtocol | None = field(default=None)

    # An error handler you can assign to manage different types of errors
    # globally or create rules within the error handler to manage how
    # errors are handled depending on database type.
    error_handlers: ErrorHandlerProtocol | list[ErrorHandlerProtocol] | None = (
        field(default=None)
    )

    # A list of handlers to process the parameters of a database execution.
    # Use cases include converting data types, validating data, cleansing
    # data, etc.
    parameter_handlers: list[ParamHandlerCallable] = field(default_factory=list)

    # A list of handlers to process the result of a database execution.
    # These handlers can process results such as filtering data from
    # certain sources, serializing data, or performing other
    # operations on the result.
    # Future version may convert this to a dictionary mapping of :type:, Callable | list[Callable]
    result_handlers: list[ResultHandlerCallable] = field(default_factory=list)

    def __post_init__(self) -> None:
        """Add in result types"""

        if not self.result_types:
            self.result_types = [
                Result,
                CursorResult[Any],
                declarative_base(),
                ScalarResult,
            ]

    @property
    def url_string(self) -> str | URL:
        """Return the URL if assigned, otherwise return the param URL"""
        return self.db.url if self.db.url else self.param_url

    @property
    def param_url(self) -> URL | str:
        """Return the connection URL built from connection parameters."""
        if self.db.url_params:
            return URL.create(**self.db.url_params.model_dump())
        return ""

    @property
    def url_base(self) -> str:
        """The base string for the connection uri"""
        return "{drivername}://{username}:{password}@{host}:{port}/{database}"

    @property
    def schema(self) -> str:
        """Get the schema name from the configuration"""
        s = self.db.statements
        if s and s.schema_name:
            return f"{s.schema_name}."
        return ""


class ConnectionBase(ConnectionConfig):
    """
    A base class that provides the base database attributes and initializes
    the configuration.
    """  # A mapping of handlers passed in as a dictionary.

    def __init__(
        self,
        config: DatabaseEngineConfig | None = None,
        engine_key: str | None = None,
        handlers: HandlerMapping | None = None,
        **kwargs,
    ) -> None:
        """
        Initialize the configuration object.

        Validate the type of configuration model being used. If no
        db_config is provided. Attempt to find a *elixir*.yaml/yml file
        within the project. Assign handlers and select the active engine
        configuration to be used if the configuration has multiple engines.

        Raises:
            ConfigNotFoundError: If no configuration is found, or a 'engine_key'
                key was provided, but no configuration matched that key.
        """

        # If a configuration is not provided, attempt to find it. The yaml
        # file must be prefixed with elixir. (e.g. myelixirdb.yaml, myelixirdb.yml)
        config = config or load_config()
        if not config:
            raise ElixirFileNotFoundError(
                "There were no configurations found "
                "in the project that matched `*elixir*.yaml`. "
            )

        # If the provided configuration is a dictionary, then validate it.
        if isinstance(config, dict):
            # Coerce it to a regular dict if it is a TypedDict
            config = dict(config)
            try:
                if config.get("engines"):
                    config = EngineManager(**config)
                else:
                    # If the config does not have `engines` as a key,
                    # assume it is a single Engine configuration. If it is not
                    # an EngineModel, it will fail validation.
                    config = EngineModel(**config)
            except ValidationError as e:
                print_and_raise_validation_errors(e)

        # The config should be a pydantic model at this point.
        if isinstance(config, EngineManager):
            # Set self.config, but only if it is not already set.
            self.config = config if not self.config else self.config
            engine_key = engine_key or config.default_engine_key or None
            if engine_key:
                db_config = config.engines.get(engine_key)
                if not db_config:
                    raise EngineKeyNotFoundError(engine_key=engine_key)
            # If there is only one configuration, then load it.
            elif len(config.engines) == 1:
                engine_key = engine_key or next(iter(config.engines.keys()))
                db_config = config.engines.get(engine_key)
            else:
                msg = (
                    "An engine_key was not provided and no default configurations "
                    "were found. Please see the quick start guide in the readme "
                    "for information on setting up your engines. "
                )
                keys = list(config.engines.keys())
                if keys:
                    msg += f"Available configurations: {keys}"
                raise EngineKeyNotDefinedError(msg)

        elif isinstance(config, EngineModel):
            # The config must be a EngineModel if validation has passed and
            # it is not a EngineManager model.
            db_config = config
        else:
            raise TypeError("Invalid configuration type. ")

        super().__init__(db=db_config, engine_key=engine_key, **kwargs)

        # Assign to class attribute
        if not self.config:  # Ensure it doesn't overwrite the class config
            self.config = config

        if not self.result_types:
            self.result_types = [Result, CursorResult[Any]]

        # Set the processing handlers.
        if handlers:
            self.set_handlers(handlers)

    @property
    def handler_names(self):
        """
        Return the names of the handlers available to be set.

        result_handlers:
            Process the results of a raw/orm query based on self.result_types
            that are set. Use cases could include converting ORM result objects or
            CursorResults to a list.

            See :class:'ConnectionnBase.result_types'
        parameter_handlers:
            Processes the parameters of a raw statement. This can be for
            cleansing the parameters or ensuring the params are compatible
            with the database charset, etc.

        error_handlers:
            Manage different error types within the instance instead of
            deploying handling for each function/method that may use the instance.
            Reduces redundancy in handling and can also be used for logging
            to a different source/engine.

        """
        return ["result_handlers", "parameter_handlers", "error_handlers"]

    def set_handlers(self, handlers: HandlerMapping) -> None:
        """Sets the handlers for the database operations."""

        for handler_name, _handler in handlers.items():
            # Check if the attribute exists in __slots__
            current_handler = getattr(self, handler_name)

            if handler_name not in self.handler_names:
                raise AttributeError(
                    f"Invalid handler name: {handler_name}. "
                    f"Valid handlers are: {self.handler_names}"
                )

            if isinstance(current_handler, list):
                # Append to the list if it's already a list
                if isinstance(_handler, list):
                    current_handler.extend(_handler)
                else:
                    current_handler.append(_handler)
            else:
                # If the dataclass did not set the default factory, it
                # should be set directly and not as a list.
                setattr(self, handler_name, _handler)

    def add_result_type(self, result_type: Any) -> None:
        """
        Add a new result type to the result types that are checked when
        determining if result_handlers should process the specified result
        type.
        >> db.add_result_type(SQLModel)

        if self.result_handlers and (
            isinstance(result.__class__, DeclarativeMeta)
            or isinstance(result, (self.result_types))
        ):
            return self._process_handler("result_handler", result)

        Args:
            result_type: The type to add for result validation

        Raises:
            TypeError: If validate=True and result_type is not a valid type

        Warns:
            UserWarning: If result_type already exists in result_types list
        """
        if not isinstance(result_type, type):
            raise TypeError(f"Expected a type, got {type(result_type)}")

        if result_type in self.result_types:
            warnings.warn(
                f"Result type {result_type.__name__} already exists.",
                UserWarning,
                stacklevel=2,
            )
            return

        self.result_types.append(result_type)



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from pydantic_core import core_schema
from sqlalchemy.engine import Engine


if TYPE_CHECKING:
    from pydantic.json_schema import JsonSchemaValue


ConnectArgsType = dict[str, Any]
ExecutionOptionsType = dict[str, Any]
JsonDeserializerType = Callable[[str], Any]
JsonSerializerType = Callable[[Any], str]
PluginListType = list[str]


class EngineSchema:
    @classmethod
    def __get_pydantic_core_schema__(  # noqa: PLW3201
        cls,
        source_type: type[Engine],
        handler: Callable[[type[Any]], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, v: Any) -> Engine:
        if not isinstance(v, Engine):
            raise TypeError("value is not an Engine instance")
        return v

    @classmethod
    def __get_pydantic_json_schema__(  # noqa: PLW3201
        cls,
        schema: core_schema.CoreSchema,
        handler: Callable[[core_schema.CoreSchema], JsonSchemaValue],
    ) -> JsonSchemaValue:
        return {
            "type": "sqlalchemy.engine.Engine",
            "title": "Engine",
            "description": "SQLAlchemy Engine object as a string",
            "examples": ["Engine(postgresql//user:pass@localhost/dbname)"],
        }


class CallableSchema:
    @classmethod
    def __get_pydantic_core_schema__(  # noqa: PLW3201
        cls,
        source_type: Callable,
        handler: Callable[[type[Any]], core_schema.CoreSchema],
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, v: Any) -> Callable:
        if not callable(v):
            raise TypeError("value is not a callable")
        return v

    @classmethod
    def __get_pydantic_json_schema__(  # noqa: PLW3201
        cls,
        schema: core_schema.CoreSchema,
        handler: Callable[[core_schema.CoreSchema], JsonSchemaValue],
    ) -> JsonSchemaValue:
        return {
            "type": "callable",
            "title": "Callable",
            "description": "Callable object as a string",
            "examples": ["Callable(lambda x: x + 1)"],
        }



"""
Base Models
"""

from __future__ import annotations

from pydantic import BaseModel
from pydantic import ConfigDict


class StrictModel(BaseModel):
    """
    Set the default configuration for the BaseModel
    """

    # We allow here because we want to provide the validation error at the end
    model_config = ConfigDict(
        extra="forbid",
        validate_assignment=True,
        arbitrary_types_allowed=True,
        str_strip_whitespace=True,
        use_enum_values=True,
    )



"""
This module provides utility functions for working with SQL queries and
database configurations.
"""

from __future__ import annotations

import re
from typing import TYPE_CHECKING
from typing import Any
import sqlglot
from sqlglot import exp


if TYPE_CHECKING:
    from elixirdb.types import DialectName
    from elixirdb.types import HandlerSequence
    from elixirdb.types import SchemaName
    from elixirdb.types import SQLStatement
    from elixirdb.types import TableName
    from elixirdb.types import _CoreAnyExecuteParams


def is_stored_procedure(query: SQLStatement) -> bool:
    """
    Determine if a given SQL query string is a stored procedure call.

    Args:
        query: The SQL query string to be checked.

    Returns:
        True if the query is identified as a stored procedure call.

    Examples:
        >>> is_stored_procedure("EXEC my_stored_procedure")
        True
        >>> is_stored_procedure("SELECT * FROM my_table")
        False
    """
    query = query.strip().lower()

    patterns = [
        r"^exec\s",  # SQL Server, Sybase
        r"^execute\s",  # SQL Server, Sybase (alternative)
        r"^call\s",  # MySQL, postgres, Oracle
        r"^begin\s",  # PL/SQL block in Oracle
        r"^declare\s",  # PL/SQL anonymous block
    ]

    return any(re.match(pattern, query) for pattern in patterns)


TEMP_TABLE_PREFIXES = {
    "mssql": ["#", "tempdb..#"],
    "mysql": ["tmp_", "temp_"],
    "mariadb": ["tmp_", "temp_"],
    "sqlite": ["temp.", "temp_"],
    "postgres": ["pg_temp."],
    "oracle": ["temp_", "ora$ptt_", "sess$", "sys_temp_"],
}


def is_temp_table(table_name: TableName, dialect: DialectName) -> bool:
    """
    Checks if a table name represents a temporary table based on the dialect.

    Args:
        table_name: The name of the table to check.
        dialect: The SQL dialect to consider.

    Returns:
        True if the table name suggests a temporary table, False otherwise.
    """
    table_name = table_name.lower()
    prefixes = TEMP_TABLE_PREFIXES.get(dialect, [])
    return any(table_name.startswith(prefix) for prefix in prefixes)


def has_sorting(sql: SQLStatement) -> bool:
    """
    Check if an SQL query contains an "ORDER BY" clause outside subqueries.

    Args:
        sql: The SQL query string to be checked.

    Returns:
        True if the SQL query contains an "ORDER BY" clause.
    """
    sql = re.sub(r"\([^()]*\)", "", sql)
    order_by_pattern = re.compile(r"\bORDER\s+BY\b", re.IGNORECASE)
    return bool(order_by_pattern.search(sql))


def has_paging(sql: SQLStatement) -> bool:
    """
    Detect paging with limits and offsets in an SQL query.

    Args:
        sql: The SQL query to analyze.

    Returns:
        True if paging is detected.
    """
    paging_patterns = [
        r"\bLIMIT\s+(\d+)\s*(?:OFFSET\s+(\d+))?\b",
        r"\bLIMIT\s+(\d+)\b(?:\s+OFFSET\s+(\d+)\b)?",
        r"\bTOP\s+(\d+)\b(?:\s+OFFSET\s+(\d+)\b)?",
        r"\bROWNUM\s*\b(?:\s*\bBETWEEN\s*\d+\s*AND\s*\d+\b|\s*\b<=?\s*\d+\b)",
        r"\bLIMIT\s+(\d+)\s*(?:OFFSET\s+(\d+))?\b",
    ]

    return any(
        re.search(pattern, sql, re.IGNORECASE) for pattern in paging_patterns
    )


def build_sql_proc_params(params: _CoreAnyExecuteParams) -> str:
    """Formats parameters for SQL stored procedure query."""
    return f"({', '.join(f':{key}' for key in params)})" if params else ""


def process_params(
    params: dict[str, str] | tuple[tuple[str, str], ...], handlers: HandlerSequence
) -> dict[str, str] | tuple[tuple[str, str], ...]:
    """
    Preprocess parameters before passing them to the database query.

    Args:
        params: Parameters to process (dict or tuple of key-value pairs)
        handlers: List of processor functions to apply to parameter values

    Returns:
        Processed parameters as dictionary or tuple
    """
    if isinstance(params, dict):
        return {key: h(value) for key, value in params.items() for h in handlers}

    return tuple((key, h(value)) for key, value in params for h in handlers)


def return_mapped_dialect(dialect: DialectName) -> str:
    """Return mapped dialect for sqlglot"""
    dialect_map = {
        "mysql": "mysql",
        "mariadb": "mysql",
        "postgres": "postgres",
        "sqlite": "sqlite",
        "oracle": "oracle",
        "msql": "sqlserver",
    }
    return dialect_map.get(dialect, "")


def apply_schema_prefix(
    node: exp.Expression,
    schema_prefix: SchemaName,
    cte_names: set[str],
    dialect: DialectName,
) -> None:
    """
    Apply schema prefix to table nodes within SQL AST.

    Args:
        node: Current node in SQL AST
        schema_prefix: Schema prefix to apply
        cte_names: Set of CTE names to exclude from prefixing
        dialect: SQL dialect in use
    """
    match node:
        case exp.Table():
            table_name = node.this.name.lower()
            if table_name not in cte_names and not is_temp_table(
                table_name, dialect
            ):
                node.set("db", exp.Identifier(this=schema_prefix))

    for child in node.args.values():
        match child:
            case exp.Expression():
                apply_schema_prefix(child, schema_prefix, cte_names, dialect)
            case list():
                for item in child:
                    if isinstance(item, exp.Expression):
                        apply_schema_prefix(item, schema_prefix, cte_names, dialect)


def apply_schema_to_statement(
    query: SQLStatement, schema_prefix: SchemaName, dialect: DialectName = "mysql"
) -> SQLStatement:
    """
    Modify SQL query by applying schema prefix to table references.

    Args:
        query: Original SQL query to modify
        schema_prefix: Schema prefix to apply
        dialect: SQL dialect in use

    Returns:
        Modified SQL query with schema prefixes applied
    """
    dialect = return_mapped_dialect(dialect)
    ast = sqlglot.parse_one(query, read=dialect)
    cte_names = {cte.alias_or_name.lower() for cte in ast.find_all(exp.CTE)}
    apply_schema_prefix(ast, schema_prefix, cte_names, dialect)
    return ast.sql(dialect=dialect)


def get_default_db_count(config: dict) -> int:
    """
    Calculate the number of default databases in the given configuration.

    Args:
        config (dict): The database configuration dictionary that matches the
        MultiEngineModel schema.

    Returns:
        int: The count of databases marked as default in the configuration.

    """
    return sum(
        1 for db in config.values() if isinstance(db, dict) and db.get("default")
    )


def is_dml_query(statement: SQLStatement) -> bool:
    """Checks if a SQL statement is a DML query (INSERT, UPDATE, DELETE)."""
    try:
        parsed = sqlglot.parse_one(statement)
        return isinstance(parsed, (exp.Insert, exp.Update, exp.Delete))
    except Exception:  # pylint: disable=broad-except
        return False


def is_list_of_type(obj: Any, type_: type, subclass: bool = False) -> bool:
    """
    Check if an object is a list where all elements are of a specific type.

    Args:
        obj (Any): The object to check.
        type_ (type): The type to check the elements against.

    Returns:
        bool: True if obj is a list of type_, otherwise False.
    """
    if not subclass:
        return isinstance(obj, list) and all(
            isinstance(item, type_) for item in obj
        )
    return isinstance(obj, list) and all(issubclass(item, type_) for item in obj)



"""
Module for parsing YAML/JSON configuration files. Support is present only for
one file, but future iterations may parse entire directories and merge them
as well as support for specific loading configurations in pyproject.toml.
"""

import json
from pathlib import Path
from typing import Any
from typing import Literal
import yaml
from pyrootutils import find_root


def scan_files(partial: str, type: str | None = None) -> list[Path]:
    """
    Recursively find files matching the given partial name.

    Args:
        partial (str): The partial string to search for in filenames.

    Returns:
        list[Path]: A list of Path objects that match the criteria.
    """
    base_dir = find_root()
    if type:
        return list(base_dir.rglob(f"*{partial}*{type}"))
    return list(base_dir.rglob(f"*{partial}*"))


def load_config(
    partial: str = "elixir", file_type: Literal["json", "yaml", "yml"] = "yaml"
) -> dict[str, Any] | None:
    """
    Load a configuration file by searching the project for JSON/YAML files
    matching the given partial name.

    Args:
        partial (str): The partial string to search for in filenames.
        file_type (Literal["json", "yaml", "yml"] | None): The file type to load.
            If not specified, all matching files are considered.

    Returns:
        dict[str, Any] | None: A dictionary of configurations if a file is found,
        otherwise None.

    Raises:
        ValueError: If more than one matching configuration file is found or if the
        file type is unsupported.
    """
    files = scan_files(partial, file_type)

    if len(files) > 1:
        raise ValueError(
            "elixir-db only supports one configuration file at the moment. "
            "Please aggregate configurations into one file."
        )
    if files:
        file = files[0]
        if file.suffix in [".yaml", ".yml"]:
            with open(file, "r", encoding="utf-8") as f:
                data = yaml.load(f, Loader=yaml.SafeLoader)
                if isinstance(data, dict):
                    return data
                else:
                    raise ValueError("YAML content not parsed to dictionary.")
        elif file.suffix == ".json":
            with open(file, "r", encoding="utf-8") as f:
                return json.load(f)
        else:
            raise ValueError(f"Unsupported file type: {file.suffix}")
    return None
